/*
	YIFF Sound server IO functions



 */

#ifndef YCLIENTIO_H
#define YCLIENTIO_H

#include <stdio.h>
#include <db.h>
#include <sys/types.h>


/*
 *	All purpose NULL YID:
 */
#define YIDNULL		0

/*
 *	Types:
 */
typedef char Boolean;
typedef unsigned long YID;
typedef double Coefficient;
typedef long YDataPosition;
typedef long YDataLength;


/*
 *	Maximums:
 */
#ifndef YMODE_NAME_MAX
# define YMODE_NAME_MAX	256
#endif

#ifndef YHOST_NAME_MAX
# define YHOST_NAME_MAX	256
#endif

#ifndef YMIXER_MAX_VALUES
# define YMIXER_MAX_VALUES	5
#endif

#define MAX_QUEUED_YIFF_EVENTS	200


/*
 *	Event type codes:
 *
 *	Note, the event codes names are prefixed YEvent!!
 *	Event structure types types are prefixed YiffEvent!!
 */
#define YEventModeChange	1	/* Server has changed modes. */
#define YEventCycleChange	2	/* Server has changed cycle. */
#define YEventDisconnect        3       /* Server disconnected connection. */
#define YEventHost		4	/* Host add or remove. */
#define YEventSoundKill         5       /* Playstack killed. */
#define YEventMixer		6	/* Mixer change. */
#define YEventSoundPlay         7       /* Playstack started. */
#define YEventShutdown		8	/* Server has shut down. */
#define YEventSync		9	/* Server has just finished syncing. */
#define YEventSystemStats	10	/* Server's system stats report. */



/*
 *	Event structures:
 */

/* Change mode event structure. */
typedef struct {

	Boolean preset;		/* Use preset? */

	/* If preset is True, these members contain right info. */
	char mode_name[YMODE_NAME_MAX];

	/* If preset is False, these members contain right info. */
	int sample_size;
	int channels;
	int sample_rate;
	int direction;
	int allow_fragmenting;
	int num_fragments;
	int fragment_size_bytes;

} YiffEventModeChange;

/* Change cycle event structure. */
typedef struct {

	long cycle_us;

} YiffEventCycleChange;

/* Disconnect event structure. */
typedef struct {

	int reason;	/* Why you were disconnected. */

} YiffEventDisconnect;

/* Host add or remove event structure. */
typedef struct {

	int op;		/* 0 for remove, 1 for add, 2 for FYI. */

	char address[YHOST_NAME_MAX];

} YiffEventHost;

/* Sound kill event structure. */
typedef struct {

	YID yid;

} YiffEventSoundKill;

/* Mixer event structure. */
typedef struct {

	int code;		/* Mixer device code. */

        double value[YMIXER_MAX_VALUES];	/* Argument values. */

} YiffEventMixer;

/* Sound play event structure. */
typedef struct {

	YID yid;

} YiffEventSoundPlay;

/* Shutdown event structure. */
typedef struct {

	int reason;	/* Why server shut down. */

} YiffEventShutdown;

/* Sync event structure. */
typedef struct {

	long cycle_ahead_us;

} YiffEventSync;

/* System stats structure. */ 
typedef struct { 

	long cycle_us;
	long compensated_cycle_us;
	int num_fragments;
	int buf_len;
	int channels;
	int sample_size;
	int sample_rate;
	long start_time;
	double cycle_load;

} YiffEventSysStats;

/* Main yiff event structure. */
typedef struct {

        int type;       /* Event type code, one of YEvent*. */

        YiffEventModeChange     mode;
        YiffEventCycleChange    cycle;
        YiffEventDisconnect     disconnect;
        YiffEventHost           host;
        YiffEventSoundKill      kill;
        YiffEventMixer          mixer;
        YiffEventSoundPlay      play;
        YiffEventShutdown       shutdown;
        YiffEventSync           sync;
        YiffEventSysStats       sysstats;

} YiffEvent;



/*
 *	Connection to YIFF server structure:
 */
typedef struct {
   
        int fd;
   
        int we_started_server;

        int queued_yevent_pos;
        YiffEvent queued_yevent[MAX_QUEUED_YIFF_EVENTS];

        YID prev_generated_yid;

} YiffConnection;





/* ***************************************************************** */

extern YiffConnection *YIFFConnect(
	char *start_arg,	/* Can be NULL for don't start if needed. */
	char *con_arg		/* Should not be NULL. */
);

extern int YIFFSetMode(
	YiffConnection *con,
	int sample_size,
	int channels,
	int sample_rate,
	int direction,
	int allow_fragmenting,
	int num_fragments,
	int fragment_size
);
extern int YIFFSetModePreset(
        YiffConnection *con,
        char *name
);
extern int YIFFSetCycle(
        YiffConnection *con,
        long usec
);
extern void YIFFSync(
	YiffConnection *con,
	Boolean block
);

extern int YIFFRequestStats(YiffConnection *con);

extern int YIFFHost(
	YiffConnection *con,
        int op,
	char *address
);
extern int YIFFSetMixer(  
        YiffConnection *con,
        int mixer_code,
        double value1,
        double value2,
        double value3,
        double value4,
        double value5
);

extern YID YIFFPlaySound(
	YiffConnection *con,
	char *path,		/* Path on disk on server's computer. */
	YDataPosition pos,
	Coefficient left_volume,
	Coefficient right_volume,
	int repeats
);
extern void YIFFDestroyPlay(YiffConnection *con, YID yid);

extern void YIFFDisconnect(YiffConnection *con, Boolean no_shutdown);
extern void YIFFShutdown(YiffConnection *con);

extern int YIFFGetNextEvent(
	YiffConnection *con,
	YiffEvent *event,
	Boolean block
);

#endif	/* YCLIENTIO_H */
