/*
	                 Audio Format Wrapper

	Functions:

	int AFWOpen(char *path, AFWDataStruct *af_data)
	int AFWLoadSegment(
	        AFWDataStruct *af_data,
	        YDataPosition position,
	        YDataLength length
	)
	void AFWClose(AFWDataStruct *af_data)



	---

	IO of audio file data regardless of format.

 */

#include <stdio.h>
#include <malloc.h>
#include <sys/stat.h>
#include <unistd.h>

#include "afw.h"



int AFWOpen(char *path, AFWDataStruct *af_data)
{
	int i, status;
	raw_data_struct *raw_data;
	wav_data_struct *wav_data;
	voc_data_struct *voc_data;
	struct stat stat_buf;


	if((path == NULL) ||
           (af_data == NULL)
	)
	    return(-1);

	/* Does file exist? */
	if(stat(path, &stat_buf))
	    return(-1);

	/* Is file a directory? */
	if(S_ISDIR(stat_buf.st_mode))
	    return(-1);


        /* ****************************************************** */
	/* Is it a WAV file? */
	if(WavIsFileWav(path) == WavSuccess)
	{
	    /* Load as a WAV file. */
	    af_data->ptr = (wav_data_struct *)calloc(
		1,
		sizeof(wav_data_struct)
	    );
	    if(af_data->ptr == NULL)
	        return(-1);
            wav_data = (wav_data_struct *)af_data->ptr;

	    status = WavReadHeader(
		path,
		wav_data
	    );
	    if(status != WavSuccess)
	    {
		free(af_data->ptr);
		af_data->ptr = NULL;

		return(-1);
	    }

            af_data->format = FileFormatWav;
	    af_data->entire_length = wav_data->data_chunk_size;
	    af_data->buffer = NULL;
	    af_data->length = 0;

	    af_data->block_align = wav_data->block_align;
	    af_data->block_length = (wav_data->bits_per_sample == 16) ?
                2 : 1;
	}
	/* ****************************************************** */
	/* Is it a VOC file? */
	else if(VocIsFileVoc(path) == VocSuccess)
	{
            /* Load as a VOC file. */
            af_data->ptr = (voc_data_struct *)calloc(
                1,
                sizeof(voc_data_struct)
            );
            if(af_data->ptr == NULL)
                return(-1);
            voc_data = (voc_data_struct *)af_data->ptr;

            status = VocReadHeader(
                path,
                voc_data
            );
            if(status != VocSuccess)
            {
                free(af_data->ptr);
                af_data->ptr = NULL;
                
                return(-1);
            }

	    af_data->format = FileFormatVoc;
            af_data->entire_length = voc_data->total_data_len;
            af_data->buffer = NULL;
            af_data->length = 0;

	    /* Get block align and length. */
            af_data->block_align = 0;
            af_data->block_length = 1;
	    for(i = 0; i < voc_data->total_datablocks; i++)
	    {
		if(voc_data->datablock[i] == NULL)
		    continue;
		if(voc_data->datablock[i]->type != VocBlockTypeSoundData)
                    continue;

                af_data->block_align = 0;
                af_data->block_length =
                    (voc_data->datablock[i]->bits == 16) ? 2 : 1;
	    }
	}
        /* ****************************************************** */
        else if(RawIsFileRaw(path) == RawSuccess) 
        {
            /* Load as a RAW file. */
            af_data->ptr = (raw_data_struct *)calloc(
                1,
                sizeof(raw_data_struct)
            );
            if(af_data->ptr == NULL)
                return(-1);
            raw_data = (raw_data_struct *)af_data->ptr;

            status = RawReadHeader(
                path,
                raw_data
            );
            if(status != RawSuccess)
            {
                free(af_data->ptr);
                af_data->ptr = NULL;

                return(-1);
            }

            af_data->format = FileFormatRaw;
            af_data->entire_length = raw_data->data_length;
            af_data->buffer = NULL;
            af_data->length = 0;

            af_data->block_align = raw_data->block_align;
            af_data->block_length = (raw_data->sample_size == 16) ?
		2 : 1;
        }
        /* ****************************************************** */
	else
	{
	    /* Unsupported file format. */
	    return(-1);
	}


	return(0);
}


/*
 *	Frees buffer on af_data, reads a segment of data from the
 *	audio file and then allocates new buffer and sets up values
 *	on af_data.
 */
int AFWLoadSegment(
        AFWDataStruct *af_data,
        YDataPosition position,
        YDataLength length
)
{
	int status;
        raw_data_struct *raw_data;
	wav_data_struct *wav_data;
	voc_data_struct *voc_data;


	if(af_data == NULL)
	    return(-1);
	if(af_data->ptr == NULL)
	    return(-1);


        /* ******************************************************** */

	/*   Do not free buffer on af_data, it will be freed by
         *   the call to the library.
	 */
	af_data->length = 0;


	/* ******************************************************** */
	/* Load audio segment. */
        switch(af_data->format)
        {
          case FileFormatRaw:
            raw_data = (raw_data_struct *)af_data->ptr;
            status = RawReadPartialData(
                raw_data,
                position,   
                length,
                0			/* Determine this later. */
            );
            if(status == RawSuccess)
            {
                af_data->buffer = raw_data->data;
                af_data->length = raw_data->data_len;
            }
            else
            {
                af_data->buffer = NULL;
            }
            break;

          case FileFormatWav:
	    wav_data = (wav_data_struct *)af_data->ptr;
	    status = WavReadPartialData(
		wav_data,
		position,		/* Position. */
		length,			/* How much to load. */
		WavByteValueShiftHalf
	    );
	    if(status == WavSuccess)
	    {
                af_data->buffer = wav_data->data;
		af_data->length = wav_data->data_len;
	    }
	    else
            {
                af_data->buffer = NULL;
            }
            break;

          case FileFormatVoc:
            voc_data = (voc_data_struct *)af_data->ptr;
	    status = VocReadPartialData(
		voc_data,
		position,
		length,
		VocByteValueShiftHalf
	    );
            if(status == VocSuccess)
            {
                af_data->buffer = voc_data->data;
                af_data->length = voc_data->data_len;
            }
            else
            {
                af_data->buffer = NULL;
            }
            break;
        
          case FileFormatMP3:
            return(-1);
            break;
 
          default:
	    return(-1);
            break;
        }


	return(0);
}


void AFWClose(AFWDataStruct *af_data)
{
	if(af_data == NULL)
	    return;

	/* Deallocate library resources and memory. */
	switch(af_data->format)
	{
	  case FileFormatRaw:
            RawDestroyData((raw_data_struct *)af_data->ptr);
	    break;

	  case FileFormatWav:
	    WavDestroyData((wav_data_struct *)af_data->ptr);
	    break;

	  case FileFormatVoc:
            VocDestroyData((voc_data_struct *)af_data->ptr);
	    break;

	  case FileFormatMP3:
	    break;

	  default:
	    break;
	}

	/* Free library structure (library does not free this). */
	free(af_data->ptr);
	af_data->ptr = NULL;

	/*   Do not free buffer, call to library destroy function
         *   has done that.
	 */
	af_data->buffer = NULL;
	af_data->length = 0;

	af_data->block_align = 0;
	af_data->block_length = 1;


	return;
}
