#include <stdio.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "raw.h"


void RawReportError(char *filename, char *reason, int how_bad);
FILE *RawOpenFile(char *filename, char *mode, off_t *size_rtn);


void RawReportError(char *filename, char *reason, int how_bad)
{
        if(filename != NULL)
        {
            fprintf(stderr, "%s: ", filename);
        }
        if(reason != NULL)
        {
            fprintf(stderr, "%s (Error level %i)", reason, how_bad);
        }
        fprintf(stderr, "\n");


        return;
}


void RawDestroyData(raw_data_struct *rd)
{
	if(rd == NULL)
	    return;

	free(rd->filename);
	rd->filename = NULL;

	rd->data_offset = 0;
	rd->data_length = 0;

        rd->block_align = 0;
        rd->sample_size = 0;
        rd->channels = 0;
        rd->sample_rate = 0;

	free(rd->data);
	rd->data = NULL;

	rd->data_len = 0;

	return;
}


FILE *RawOpenFile(char *filename, char *mode, off_t *size_rtn)
{
        static FILE *fp;
        static struct stat stat_buf;


        if(filename == NULL)
        {
            fprintf(stderr, "Cannot open file with no name.\n");
            *size_rtn = 0;
            return(NULL);
        }
        if(mode == NULL)
        {
            fprintf(stderr, "%s: Open mode not givin.\n", filename);
            *size_rtn = 0;
            return(NULL);
        }

        if(stat(filename, &stat_buf))
        {
            fprintf(stderr, "%s: No such file.\n", filename);
            *size_rtn = 0;
            return(NULL);
        }
        *size_rtn = stat_buf.st_size;

        fp = fopen(filename, mode);
        if(fp == NULL)
        {
            fprintf(stderr, "%s: Cannot open.\n", filename);
            *size_rtn = 0;
            return(NULL);
        }


        return(fp);
}


int RawIsFileRaw(char *filename)
{
        static FILE *fp;
        static off_t filesize;


        /* Open file. */
        fp = RawOpenFile(filename, "rb", &filesize);
        if(fp == NULL)
            return(RawErrorNoAccess);
        if(filesize == 0)
            return(RawErrorNoAccess);

/* What do we check? */


        /* Close file. */
        fclose(fp);
	fp = NULL;


        return(RawSuccess); 
}


int RawReadHeader(char *filename, raw_data_struct *rd)
{
	static int len;
        static FILE *fp;
        static off_t filesize;


        /* Error checks. */
        if(rd == NULL)
            return(RawErrorNoBuffers);
        if(filename == NULL)
            return(RawErrorBadValue);


        /* Reset values. */
        RawDestroyData(rd);


        /* Open file. */
        fp = RawOpenFile(filename, "rb", &filesize);
        if(fp == NULL)
            return(RawErrorNoAccess);
        if(filesize == 0)
            return(RawErrorNoAccess);


        /* Record file name. */
	len = strlen(filename);
        rd->filename = (char *)calloc(1, (len + 1) * sizeof(char));
        strcpy(rd->filename, filename);

/* Check for offset? */
	/* Get offset. */
	rd->data_offset = 0;

	/* Get length of data. */
	rd->data_length = (long)filesize - (long)rd->data_offset;
	if(rd->data_length < 0)
	    rd->data_length = 0;


	/* Standard values. */
	rd->block_align = 0;
	rd->sample_size = 8;
        rd->channels = 1;
	rd->sample_rate = 8000;


        /* Close file. */
        fclose(fp);
        fp = NULL;


        return(RawSuccess);
}


int RawReadPartialData(
        raw_data_struct *rd,
        long offset,            /* In bytes. */
        long max_chunk_size,    /* In bytes. */
        int read_opt
)
{
	static int i;
	static FILE *fp;
	static off_t filesize;
	static char *buf_ptr;


        /* Error checks. */
        if(rd == NULL)
            return(RawErrorNoBuffers);
        if(max_chunk_size <= 0)
            return(RawErrorBadValue);
        if(offset < 0)
            return(RawErrorBadValue);

        /* Free data if already loaded. */
        free(rd->data);
	rd->data = NULL;
        rd->data_len = 0;


	/* Sanitize values. */
	if(offset >= rd->data_length)
        {
            return(RawErrorEndOfData);
	}

	if((offset + max_chunk_size) >= rd->data_length)
	{
            max_chunk_size = rd->data_length - (offset + 1);
	}

        if(max_chunk_size <= 0)
        {
            return(RawErrorEndOfData);
        }


        /* ********************************************************* */

        /* Open file. */
        fp = RawOpenFile(rd->filename, "rb", &filesize);
        if(fp == NULL)
            return(RawErrorNoAccess);
        if(filesize == 0)
            return(RawErrorNoAccess);

        /* Allocate memory. */
        rd->data = (char *)calloc(1, max_chunk_size * sizeof(char));
        if(rd->data == NULL)
        {
            fclose(fp);
            fp = NULL;
            return(RawErrorNoBuffers);
        }
        rd->data_len = max_chunk_size;


        /* Position file pointer. */
        fseek(fp, rd->data_offset + offset, 0);

        /* Read the data. */
	buf_ptr = rd->data;
        switch(read_opt)
        {
          /* Shift byte value by - 128. */
          case 1:
            for(i = 0; i < max_chunk_size; i++)
                *buf_ptr++ = (char)(fgetc(fp) - 128);
            break;

	  /* Get data as is. */
          default:
            for(i = 0; i < max_chunk_size; i++)
                *buf_ptr++ = (char)fgetc(fp);
            break;
        }


        /* Close the file. */
        fclose(fp);
        fp = NULL;


        return(RawSuccess);
}
