#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


#ifdef __linux__
# include <linux/soundcard.h>
#endif	/* __linux__ */

#include "yiff.h"


#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define MAX(a,b)	(((a) > (b)) ? (a) : (b))


int YSoundInit(int argc, char **argv, Recorder *recorder)
{
	int i, status;
        Sound *sound_ptr;
        Audio *audio_ptr;
#ifdef SOUND_EXPERIMENTAL
	unsigned int fragment, num_fragments, fragment_size;
	int buffer_size;
#endif	/* SOUND_EXPERIMENTAL */


	if(recorder == NULL)
            return(-1);


        /* Get pointers to substructures. */
        sound_ptr = &recorder->sound;
        audio_ptr = &recorder->audio;


	/* Reset values to defaults. */
	bzero(sound_ptr, sizeof(Sound));
	bzero(audio_ptr, sizeof(Audio));

        sound_ptr->buffer_length = 0;
        sound_ptr->buffer_content_length = 0;

        sound_ptr->device_buffer_length = 0;
        sound_ptr->device_buffer_content_length = 0;

	audio_ptr->cycle_set = option.cycle_set;

	audio_ptr->cycle.ms = 380;		/* Default value */
        audio_ptr->cycle.us = 0;		/* Default value */

        audio_ptr->compensated_cycle.ms = audio_ptr->cycle.ms;
	audio_ptr->compensated_cycle.us = audio_ptr->cycle.us;

        audio_ptr->write_ahead.ms = 0;
        audio_ptr->write_ahead.us = 0;

        audio_ptr->cumulative_latency.ms = 0;
        audio_ptr->cumulative_latency.us = 0;

        audio_ptr->sample_size = option.sample_size;
        audio_ptr->channels = option.channels;
        audio_ptr->sample_rate = option.sample_rate;
        audio_ptr->bytes_per_second = (double)option.sample_rate *
            (double)MAX((option.sample_size / 8), 1) *
            (double)(audio_ptr->cycle.ms / 1000)
	;

#ifdef SOUND_EXPERIMENTAL
        audio_ptr->allow_fragments = option.allow_fragments;
        audio_ptr->num_fragments = option.num_fragments;
        audio_ptr->fragment_size = option.fragment_size;
#endif	/* SOUND_EXPERIMENTAL */

        audio_ptr->flip_stereo = option.flip_stereo;

	audio_ptr->direction = option.direction;
        audio_ptr->fd = -1;


        /* ******************************************************** */
	/* Parse arguments. */
	for(i = 1; i < argc; i++)
	{
	    if(argv[i] == NULL)
		continue;

	}


        /* ******************************************************** */
	/* Open audio device. */
	audio_ptr->fd = open(
	    fname.device,
	    (audio_ptr->direction == AUDIO_DIRECTION_RECORD) ?
		O_RDONLY : O_WRONLY
	);
        if(audio_ptr->fd < 0)
	{
	    fprintf(stderr,
		"%s: Cannot open for %s.\n",
		fname.device,
		(audio_ptr->direction == AUDIO_DIRECTION_RECORD) ?
		    "recording" : "playing"
	    );

	    YSoundShutdown(recorder);
	    return(-1);
	}


#ifdef SOUND_EXPERIMENTAL
        /* ********************************************************** */
        /*      
         *  There is a new ioctl called SNDCTL_DSP_SETFRAGMENT. It
         *  accepts an int parameter which has format 0x00nn00ss where
         *  the nn is max number of buffer fragments (between 0x02 and 0xff)
         *  and the ss gives indirectly the size of a buffer fragment
         *  (fragment_size = (1 << ss)). Valid sizes are between
         *  (ss=0x07 -> 128 bytes and ss=0x11 (17 dec) -> 128k).   
         *   
         *      Linux documentation suggestion: 0x00020009
         *      XBlast Sound Server: 0x0004000a
         */
        if(audio_ptr->allow_fragments)
        {
	    num_fragments = audio_ptr->num_fragments;
	    fragment_size = audio_ptr->fragment_size;

	    /* Pack fragment argument. */
            fragment = (num_fragments << 16) | fragment_size;

            status = ioctl(
		audio_ptr->fd,
		SNDCTL_DSP_SETFRAGMENT,
		&fragment
	    );
	    if(status == -1)
	    {
		fprintf(stderr,
     "%s: Warning: Cannot set buffer fragment configuration to 0x%.8x\n",
		    fname.device, fragment
		);
	    }
        }
#endif  /* SOUND_EXPERIMENTAL */ 


        /* Set sample_size. */
        status = ioctl(
	    audio_ptr->fd,
	    SNDCTL_DSP_SAMPLESIZE,
	    &audio_ptr->sample_size
	);
	if(status == -1)
        {
            fprintf(stderr,
                "%s: Warning: Cannot set sample size to %i bits.\n",
                fname.device, audio_ptr->sample_size
            );
        }

        /* Set sample_rate. */
	status = ioctl(
	    audio_ptr->fd,
            SNDCTL_DSP_SPEED,	/* aka SOUND_PCM_WRITE_RATE. */
	    &audio_ptr->sample_rate
	);
	if(status == -1)
        {
            fprintf(stderr,
                "%s: Warning: Cannot set sample rate to %i Hz.\n",
                fname.device, audio_ptr->sample_rate
            );
        }

	/* Set channels. */
	status = ioctl(
            audio_ptr->fd,
	    SNDCTL_DSP_CHANNELS,
	    &audio_ptr->channels
	);
	if(status == -1)
	{
            fprintf(stderr,
                "%s: Warning: Cannot set channels to %i.\n",
                fname.device, audio_ptr->channels
            );
	}

#ifdef SOUND_EXPERIMENTAL
        /*
         *   Get exact buffer size demanded by device per cycle.
	 *
         *   All write()s to the sound device must have a buffer of
         *   this exactly this size.   SNDCTL_DSP_GETBLKSIZE must
	 *   be called after SNDCTL_DSP_SETFRAGMENT.
         */
        status = ioctl(
	    audio_ptr->fd,
	    SNDCTL_DSP_GETBLKSIZE,
	    &buffer_size
	);
	if(status == -1)
	{
            fprintf(stderr,
                "%s: Warning: Cannot get buffer size.\n",
                fname.device
            );
	}

        sound_ptr->buffer_length = buffer_size;
        sound_ptr->device_buffer_length = buffer_size;


	/*   Calculate theoretical cycle (provided by hardware).
	 *   Warning: This value is rarly correct and should be
	 *   set manually.
	 */
	if(audio_ptr->cycle_set == CYCLE_SET_HARDWARE)
	{
            audio_ptr->cycle.ms =
                (int)sound_ptr->buffer_length *
                1000 / (int)audio_ptr->sample_rate /
	        MAX(((int)audio_ptr->sample_size / 8), 1);
            audio_ptr->cycle.us =
                (double)sound_ptr->buffer_length /
                (double)audio_ptr->sample_rate * 1000000 /
                MAX(((long)audio_ptr->sample_size / 8), 1);
	    audio_ptr->cycle.us = audio_ptr->cycle.us -
	        (audio_ptr->cycle.ms * 1000);

            audio_ptr->compensated_cycle.ms = audio_ptr->cycle.ms;
            audio_ptr->compensated_cycle.us = audio_ptr->cycle.us;
	}

	/* Not sure what this is... dividess fragment segments? */
/*
        status = 4;
        ioctl(audio_ptr->fd, SOUND_PCM_SUBDIVIDE, &status);
 */

#endif	/* SOUND_EXPERIMENTAL */


	/* ********************************************************** */
        /* Allocate buffers. */
	sound_ptr->buffer = (SoundBuffer *)calloc(
	    1,
	    sound_ptr->buffer_length * sizeof(SoundBuffer)
	);
	if(sound_ptr->buffer == NULL)
	{
            YSoundShutdown(recorder);
	    return(-1);
	}

        sound_ptr->device_buffer = (SoundBuffer *)calloc(  
            1,
            sound_ptr->device_buffer_length * sizeof(SoundBuffer)
        );
        if(sound_ptr->device_buffer == NULL)
        {
            YSoundShutdown(recorder);
            return(-1);
        }


	/* Caliberate cycle (only if cycle is to be set by program). */
	if(audio_ptr->cycle_set == CYCLE_SET_PROGRAM)
	{
	    YSoundCaliberateCycle(recorder);
	}

/*
fprintf(stderr, "Cycle: %i.%i (comp %i.%i)  Buffer: %i bytes\n",
 audio_ptr->cycle.ms,
 audio_ptr->cycle.us,
 audio_ptr->compensated_cycle.ms,
 audio_ptr->compensated_cycle.us,
 sound_ptr->device_buffer_length
);

fprintf(stderr, "Channels: %i  Bits: %i  Sample Rate: %i\n",
 audio_ptr->channels,
 audio_ptr->sample_size,
 audio_ptr->sample_rate
);   
*/


        return(0);
}



/*
 *	Attempts to caliberate the right audio.cycle and
 *	audio.compensated_cycle for the sound device.
 *
 *	Consider this (program calculating cycle).
 */
void YSoundCaliberateCycle(Recorder *recorder)
{
#define TOTAL_SYNC_PASSES	4

	int i, fd;
	time_t prev_delta_u, delta_u;

        Sound *sound_ptr;
        Audio *audio_ptr;
	YTime start_time, end_time;

	SoundBuffer *buf;
	YDataLength buf_len;


	if(recorder == NULL)
	    return;

        /* Get pointers to substructures. */
        sound_ptr = &recorder->sound; 
        audio_ptr = &recorder->audio; 

	/* Not allow program to set cycle? */
	if(audio_ptr->cycle_set != CYCLE_SET_PROGRAM)
	    return;

	if(sound_ptr->device_buffer == NULL)
	    return;
	if(recorder->audio.fd < 0)
	    return;

	fd = recorder->audio.fd;
	buf = sound_ptr->device_buffer;
	buf_len = sound_ptr->device_buffer_length;


	/* Reset actual buffer for sound device. */
	memset(buf, (SoundBuffer)0x00, buf_len);


	/* ****************************************************** */
	/* Begin syncing. */

	/*   Sync in case audio is still playing.  Otherwise
	 *   the average cycle calculated is garbage.
	 */
        ioctl(recorder->audio.fd, SNDCTL_DSP_SYNC, 0);

	/*   Write and sync a couple of times, get average cycle
	 *   value.
	 */
        GetCurrentTime(&start_time);
        write(fd, buf, buf_len);
        ioctl(recorder->audio.fd, SNDCTL_DSP_SYNC, 0);
        GetCurrentTime(&end_time);

        delta_u = MAX(((end_time.ms * 1000) + end_time.us) -
                      ((start_time.ms * 1000) + start_time.us), 0
                  );
	prev_delta_u = delta_u;

	for(i = 0; i < TOTAL_SYNC_PASSES; i++)
	{
	    GetCurrentTime(&start_time);
            write(fd, buf, buf_len);
            ioctl(recorder->audio.fd, SNDCTL_DSP_SYNC, 0);
	    GetCurrentTime(&end_time);

	    delta_u = MAX(((end_time.ms * 1000) + end_time.us) -
                          ((start_time.ms * 1000) + start_time.us), 0
                      );

	    /* Average it. */
	    delta_u = (prev_delta_u + delta_u) / 2;
	    prev_delta_u = delta_u;
	}

	audio_ptr->cycle.ms = delta_u / 1000;
	audio_ptr->cycle.us = delta_u % 1000;

        audio_ptr->compensated_cycle.ms = audio_ptr->cycle.ms;
        audio_ptr->compensated_cycle.us = audio_ptr->cycle.us;


	return;
}


/*
 *	Copys the midway buffer to the sound buffer and attempts
 *	to have the sound device play it.  Afterwards the midway
 *	buffer's contents are reset.
 */
int YSoundPlayBuffer(Recorder *recorder)
{
	int bytes_written;
	Sound *sound_ptr;
        Audio *audio_ptr;


        if(recorder == NULL)
            return(-1);

        /* Get pointers to substructures. */
        sound_ptr = &recorder->sound;
        audio_ptr = &recorder->audio;


	if(audio_ptr->fd < 0)
	    return(-1);

	if((sound_ptr->buffer == NULL) ||
           (sound_ptr->device_buffer == NULL)
	)
	    return(-1);

	/* **************************************************** */
	/* Copy buffer to device_buffer. */

        memcpy(
            sound_ptr->device_buffer,		/* target. */
	    sound_ptr->buffer,			/* source. */
	    MIN(sound_ptr->buffer_length,
	        sound_ptr->device_buffer_length
	    )
	);

	/* Write to audio device. */
	bytes_written = write(
	    audio_ptr->fd,
	    sound_ptr->device_buffer,
	    sound_ptr->device_buffer_length
	);


	/* Clear midway buffer. */
        memset(
	    sound_ptr->buffer,
	    (SoundBuffer)0x00,
	    sound_ptr->buffer_length
	);


	return(0);
}


void YSoundSync(Recorder *recorder, int options)
{
        if(recorder == NULL)
            return;

	if(recorder->audio.fd > -1)
	{
            /* End any current playback or recording. */
            ioctl(recorder->audio.fd, SNDCTL_DSP_SYNC, 0);
	}


	return;
}


void YSoundShutdown(Recorder *recorder)
{
	Sound *sound_ptr;
	Audio *audio_ptr;


	if(recorder == NULL)
	    return;


	/* Get pointers to substructures. */
	sound_ptr = &recorder->sound;
	audio_ptr = &recorder->audio;


        if(audio_ptr->fd > -1)
	{
            /* Wait for sound to finish playing. */
            ioctl(audio_ptr->fd, SNDCTL_DSP_SYNC, 0);

            close(audio_ptr->fd);
            audio_ptr->fd = -1;
	}


	/* *********************************************************** */

	/* Free allocated substructures. */
	free(sound_ptr->buffer);
	sound_ptr->buffer = NULL;
	sound_ptr->buffer_length = 0;
	sound_ptr->buffer_content_length = 0;

        free(sound_ptr->buffer);
        sound_ptr->device_buffer = NULL;
        sound_ptr->device_buffer_length = 0;
        sound_ptr->device_buffer_content_length = 0;

	bzero(&audio_ptr->cycle, sizeof(YDeltaTime));
        bzero(&audio_ptr->compensated_cycle, sizeof(YDeltaTime));

        bzero(&audio_ptr->write_ahead, sizeof(YDeltaTime));
        bzero(&audio_ptr->cumulative_latency, sizeof(YDeltaTime));

	audio_ptr->sample_size = 0;
        audio_ptr->channels = 0;
        audio_ptr->sample_rate = 0;
        audio_ptr->bytes_per_second = 0;

#ifdef SOUND_EXPERIMENTAL
        audio_ptr->allow_fragments = False;
        audio_ptr->num_fragments = 0;
        audio_ptr->fragment_size = 0;
#endif /* SOUND_EXPERIMENTAL */

        audio_ptr->flip_stereo = False;
	audio_ptr->fd = -1;

/*
fprintf(stderr, "Sound shutdown.\n");
*/

	return;
}
