/*
                     YIFF Client side to Server Communications

	Functions:

	YiffConnection *YIFFConnect(
		char *start_arg,
		char *con_arg
	)

	int YIFFSetMode(  
	        YiffConnection *con,
	        int sample_size,
	        int channels,
	        int sample_rate,
		int direction,
	        int allow_fragmenting,
	        int num_fragments,
	        int fragment_size         In bytes
	)
	int YIFFSetModePreset(
	        YiffConnection *con,
	        char *name
	)
	int YIFFSetCycle(
		YiffConnection *con,
                long usec
        )
	void YIFFSync(
		YiffConnection *con,
		Boolean block
	)
        int YIFFRequestStats(
                YiffConnection *con
        )
	int YIFFHost(
	        YiffConnection *con,
	        int op,
	        char *address
	)
	int YIFFSetMixer(
		YiffConnection *con,
		int mixer_code,
		double value1,
		double value2,
		double value3,
		double value4,
		double value5
	)

	YID YIFFPlaySound(
	        YiffConnection *con,
	        char *path,
	        YDataPosition pos,
	        Coefficient left_volume,
	        Coefficient right_volume,
		int repeats
	)
	void YIFFDestroyPlay(YiffConnection *con, YID yid)

	void YIFFDisconnect(YiffConnection *con, Boolean no_shutdown)
	void YIFFShutdown(YiffConnection *con)

	int YIFFGetNextEvent(
	        YiffConnection *con,
	        YiffEvent *event,
	        Boolean block
	)

	---

 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <db.h>
#include <errno.h>
extern int h_errno;		/* For netdb.h */
#include <time.h>
#include <sys/time.h>
#include <signal.h>
#include <unistd.h>
#include <ctype.h>

#define _USE_BSD
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>

#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>          /* For inet_ntoa() */
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "../include/ynetwork.h"
#include "../include/yclientio.h"


#ifndef False
# define False	0
#endif

#ifndef True
# define True	1
#endif


/*
 *	Path and name maximums:
 */
#ifndef PATH_MAX
# define PATH_MAX       1024
#endif

#ifndef NAME_MAX
# define NAME_MAX       256
#endif


/*
 *	Recieve buffer length:
 *
 *	For YIFFGetNextEvent().
 */
#ifndef RECV_BUF_LEN
# define RECV_BUF_LEN	8000
#endif

/*
 *	Connect retries:
 */
#define YIFF_CONNECT_RETRIES	5

/*
 *	Default connection argument:
 */
#define DEF_CONNECT_ARG		"127.0.0.1:9433"

/*
 *	Read and write timeouts:
 *
 *	In seconds.
 */
#define YIFF_WRITE_TIMEOUT_SEC		1
#define YIFF_READ_TIMEOUT_SEC		2


/*
 *	Local functions:
 */
char **YIFF_STR_EXPLODE(char *str, int *n);
char **YIFF_STR_CHAR_EXPLODE(char *str, char c, int *n);

void YIFF_CHILD_KILL(int s);
char *YIFF_GET_PARENT(char *path);
char **YIFF_EXPLODE_CMD(char *cmd, int *strc);
int YIFF_START_SERVER(char *start_arg);
int YIFF_CONNECT_TO(char *address, int port);
int YIFF_SEND_DATA(int fd, char *buf);
void YIFF_COPY_EVENTS(YiffEvent *tar, YiffEvent *src);


/* ******************************************************************* */

/*
 *      Explode string by white space, returns dynamically allocated
 *      array of strings.
 */
char **YIFF_STR_EXPLODE(char *str, int *n)
{
        char **ret;
        char *head, *tail;
        int num;
        int len;


        /* return NULL if given an empty string */
        if (!str)
                return NULL;

        num = 0;
        head = tail = str;
        ret = NULL;

        /* iterate while we're inside the string */
        while (*head)
        {
                /* skip to next non-blank character */
                while (isblank(*head))
                        head++;

                /* find next blank character */
                tail = head;
                while (*tail && !isblank(*tail))
                        tail++;

                /* determine length of substring */
                len = (tail - head);

                /* add string to list */
                num++;
                ret = (char **)realloc(ret, num*sizeof(char *));
                ret[num - 1] = (char *)malloc(len + 1);

                /* copy string over and null-terminate */
                strncpy (ret[num - 1], head, len);
                ret[num-1][len] = '\0';

                /* get ready for next substring */
                head = tail;
        }

        /* we're done... */
        *n = num;
        return ret;
}

/*
 *	Explode string by character, returns dynamically allocated
 *	array of strings.
 */
char **YIFF_STR_CHAR_EXPLODE(char *str, char c, int *n)
{
        char **ret;
        char *head, *tail;
        int num;
        int len;


        /* return NULL if given an empty string */
        if (!str)
                return NULL;

        num = 0;
        head = tail = str;
        ret = NULL;

        /* iterate while we're inside the string */
        while (*head && *tail)
        {
                /* find next character matching c */
                tail = head;
                while (*tail && (*tail != c))
                        tail++;

                /* determine length of substring */
                len = (tail - head);

                /* add string to list */
                num++;
                ret = (char **)realloc(ret, num * sizeof(char *));
                ret[num - 1] = (char *)malloc((len + 1) * sizeof(char));

                /* copy string over and null-terminate */
                strncpy (ret[num - 1], head, len);
                ret[num - 1][len] = '\0';

                /* get ready for next field */
                head = tail + 1;
        }

        /* we're done... */
        *n = num;
        return ret;
}



/*
 *	SIGCHLD signal handler.
 */
void YIFF_CHILD_KILL(int s)
{
        int status;

        while(wait3(&status, WNOHANG, NULL) > 0);

	return;
}

/*
 *	Get parent directory.
 */
char *YIFF_GET_PARENT(char *path)
{
        int i, len;
        char *strptr;
        static char rtnstr[PATH_MAX];


        if(path == NULL)
        {
            /* Must return a valid path, so toplevel will have to do. */
            return("/");
        }

        /* path must be absolute. */
        if(path[0] != '/')
        {
            /* Must return a valid path, so toplevel will have to do. */
            return("/");
        }


        strncpy(rtnstr, path, PATH_MAX);
        rtnstr[PATH_MAX - 1] = '\0';

        len = strlen(rtnstr);
        i = len - 1;   
        strptr = &(rtnstr[i]);

        /* Skip tailing '/' characters. */
        while(i >= 0)
        {
            if(*strptr == '/')
            {
                i--;
                strptr--;
            }
            else
            {
                break;
            }
        }

        while(i >= 0)
        {
            if(*strptr == '/')
            {
                *strptr = '\0';
                break;
            }

            i--;
            strptr--;
        }   

        if(rtnstr[0] == '\0')
            return("/");


        return(rtnstr);
}


/*
 *	Returns a dynamically allocated array of strings
 *	`exploded' from the cmd string.
 *
 *	Quotation grouping is parsed properly.
 *
 *	Calling function must free returned strings.
 */
char **YIFF_EXPLODE_CMD(char *cmd, int *strc)
{
        int x, y, z;
        int str_num;
        int src_pos;
        int len;
        char **strv;

        char skip_quotes;

char c = ' ';


        /* Error checks. */
        if(strc == NULL)
            return(NULL);
        if(cmd == NULL)
        {
            *strc = 0;  
            return(NULL);
        }


        /* ********************************************************** */
        /* Reset values. */
        src_pos = 0;
        str_num = 0;

        strv = NULL;
        *strc = 0;
             
        len = strlen(cmd);
        skip_quotes = 0;

        /* Skip initial characters. */
        while(src_pos < len)
        {
            if(cmd[src_pos] != c)
                break;

            src_pos++;
        }


        /* Begin exploding strings. */
        while(src_pos < len)
        {
            skip_quotes = 0;

            /* Get length y of this segment x. */
            x = src_pos;
            y = 0;
            while((cmd[x] != c) && (x < len))
            {
                /* Quote grouping skip. */
                if(cmd[x] == '"')
                {
                    skip_quotes = 1;
                    x++;

                    /* Seek to end quote. */
                    while(x < len)
                    {
                        if(cmd[x] == '"')
                            break;

                        x++; y++;
                    }

                    break;
                }

                x++;
                y++;
            }
            /* y is now length of this segment. */

            /* Allocate new string for this segment. */
            str_num = *strc;
            *strc += 1;
            strv = (char **)realloc(strv, *strc * sizeof(char *));
	    if(strv == NULL)
	    {
		*strc = 0;
		return(NULL);
	    }
            strv[str_num] = (char *)calloc(1, (y + 1) * sizeof(char));
            if(strv[str_num] == NULL)
            {
                return(strv);
            }

            /* Copy segment. */
            x = src_pos;
            z = 0;
            while((x < len) && (z < y))
            {
                /* Skip quotes. */
                if(cmd[x] == '"')
                {
                    x++;
                    continue;
                }

                strv[str_num][z] = cmd[x];
                z++; x++;
            }
            strv[str_num][y] = '\0';    /* Null terminate. */

            /* Seek next src_pos. */
            if(y < 1)
                src_pos += 1;
            else
                src_pos += y;
            if(skip_quotes)
            {
                while(src_pos < len)
                {
                    if(cmd[src_pos] == '"')
                    {
                        src_pos++;
                        break;
                    }
                    src_pos++;
                }
            }
            while(src_pos < len)
            {
                if(cmd[src_pos] != c)
                    break;

                src_pos++;
            }
        }


        return(strv);
}


/*
 *	Attempts to start server.
 */
int YIFF_START_SERVER(char *start_arg)
{
        int i;
        char *strptr;
        int argc;
        char **argv;
        char new_cd[PATH_MAX];

        pid_t p;

	char *cmd;


	if(start_arg == NULL)
	    return(-1);
        cmd = start_arg;


        /* Explode cmd string, argv and its strings are allocated. */
        argv = YIFF_EXPLODE_CMD(cmd, &argc);
        if((argv == NULL) || (argc <= 0))
            return(-1);

        /* Put NULL on last pointer. */
        argv = (char **)realloc(argv, (argc + 1) * sizeof(char *));
        if(argv == NULL)
            return(-1);   
        argv[argc] = NULL;


        /* Set new current dir. */
        if(argc > 0)
        {
            strptr = YIFF_GET_PARENT(argv[0]);
            if(strptr == NULL)
                strncpy(new_cd, "/", PATH_MAX);
            else
                strncpy(new_cd, strptr, PATH_MAX);
        }
        else
        {
            strncpy(new_cd, "/", PATH_MAX);
        }
        new_cd[PATH_MAX - 1] = '\0';


        /* ********************************************************* */

        /* Set SIGCHLD to signal handler to eliminate zombies. */
        signal(SIGCHLD, YIFF_CHILD_KILL);

        /* Fork off a process. */
        p = fork();
        switch(p)   
        {
          /* Forking error. */
          case -1:
            perror("fork");
            exit(1);
            break;

          /* We are the child: run the command then exit. */
          case 0:
            /* Change dir. */
            if(chdir(new_cd))
                chdir("/");

            /* Execute command and arguments. */
            execvp(argv[0], argv);

            exit(0);
            break;

          /* We are the parent: Do nothing. */
          default:
            break;
        }


        /* ********************************************************* */

        /* Free allocated string pointers. */
        for(i = 0; i < argc; i++)
        {
            free(argv[i]);
        }
        free(argv); argv = NULL;
        argc = 0;


	return(0);
}


/*
 *	Returns socket fd or -1 on error.
 */
int YIFF_CONNECT_TO(char *address, int port)
{
	int s;
        struct hostent *host;		/* Host pointer. */
        struct sockaddr_in haddr;


	if(address == NULL)
	    return(-1);


	/* Get host address data. */
	host = gethostbyname(address);
	if(host == NULL)
	    return(-1);

	/* Get a socket. */
	s = socket(AF_INET, SOCK_STREAM, 0);
	if(s < 0)
	    return(-1);

        haddr.sin_family = AF_INET;		/* In hbo. */   
        haddr.sin_port = htons(port);		/* In nbo. */
        haddr.sin_addr = *((struct in_addr *)host->h_addr);
        bzero(&haddr.sin_zero, 8);	/* Zero the rest of struct. */

	/* Connect. */
	if(
	    connect(
		s,
		(struct sockaddr *)&haddr,
                sizeof(struct sockaddr)
	    ) == -1
	)
            return(-1);

        /* Set socket to non-blocking. */
        fcntl(s, F_SETFL, O_NONBLOCK);


	return(s);
}


/*
 *	Returns bytes sent or -1 on error.
 */
int YIFF_SEND_DATA(int fd, char *buf)
{
        static struct timeval timeout;
        static fd_set writefds;
        
        static int len, bytes_written;


	if(fd < 0)
	    return(-1);
	if(buf == NULL)
	    return(-1);

        len = strlen(buf);

        timeout.tv_sec = YIFF_WRITE_TIMEOUT_SEC;
        timeout.tv_usec = 0;
        FD_ZERO(&writefds);
        FD_SET(fd, &writefds);
        select(fd + 1, NULL, &writefds, NULL, &timeout);

        if(!FD_ISSET(fd, &writefds))
            return(-1);


        bytes_written = send(fd, buf, len, 0);
	if(bytes_written < 0)          
	    return(-1);
	else
	    return(bytes_written);
}


/*
 *	Coppies the information from YiffEvent src to tar.
 *
 *	Only the information pertainant to what type of
 *	event src is will be coppied.
 */
void YIFF_COPY_EVENTS(YiffEvent *tar, YiffEvent *src)
{
	int i;


	if((tar == NULL) ||
           (src == NULL)
	)
	    return;


	/* Common. */
	tar->type = src->type;


	/* Type specific. */
	switch(src->type)
	{
	  case YEventModeChange:
            tar->mode.preset = src->mode.preset;

            strncpy(
                tar->mode.mode_name,
                src->mode.mode_name,
                YMODE_NAME_MAX
            );
            tar->mode.mode_name[YMODE_NAME_MAX - 1] = '\0';

	    tar->mode.sample_size = src->mode.sample_size;
            tar->mode.channels = src->mode.channels;
            tar->mode.sample_rate = src->mode.sample_rate;
            tar->mode.direction = src->mode.direction;
            tar->mode.allow_fragmenting = src->mode.allow_fragmenting;
            tar->mode.num_fragments = src->mode.num_fragments;
            tar->mode.fragment_size_bytes = src->mode.fragment_size_bytes;
            break;

	  case YEventCycleChange:
            tar->cycle.cycle_us = src->cycle.cycle_us;
            break;

	  case YEventDisconnect:
            tar->disconnect.reason = src->disconnect.reason;
            break;

          case YEventHost:
            tar->host.op = src->host.op;
	    strncpy(
		tar->host.address,
		src->host.address,
		YHOST_NAME_MAX
	    );
	    tar->host.address[YHOST_NAME_MAX - 1] = '\0';
            break;

          case YEventSoundKill:
            tar->kill.yid = src->kill.yid;
            break;

          case YEventMixer:
            tar->mixer.code = src->mixer.code;
	    for(i = 0; i < YMIXER_MAX_VALUES; i++)
                tar->mixer.value[i] = src->mixer.value[i];
            break;

          case YEventSoundPlay:
            tar->play.yid = src->play.yid;
            break;

          case YEventShutdown:
            tar->shutdown.reason = src->shutdown.reason;
            break;

          case YEventSync:
            tar->sync.cycle_ahead_us = src->sync.cycle_ahead_us;
            break;

          case YEventSystemStats:
            tar->sysstats.cycle_us = src->sysstats.cycle_us;
            tar->sysstats.compensated_cycle_us =
		src->sysstats.compensated_cycle_us;
            tar->sysstats.num_fragments = src->sysstats.num_fragments;
            tar->sysstats.buf_len = src->sysstats.buf_len;
            tar->sysstats.channels = src->sysstats.channels;
            tar->sysstats.sample_size = src->sysstats.sample_size;
            tar->sysstats.sample_rate = src->sysstats.sample_rate;
            tar->sysstats.start_time = src->sysstats.start_time;
            tar->sysstats.cycle_load = src->sysstats.cycle_load;
	    break;

	  default:
	    fprintf(stderr,
		"YIFF_COPY_EVENTS(): Unsupported event type %i.\n",
		src->type
	    );
	    break;
	}

	return;
}


/*
 *	Attempts to connect to the sound server at the address
 *	and port contained in con_arg.
 *
 *	con_arg is a null terminated string of the format
 *	"address:port", ie "127.0.0.1:1234".
 *
 *	If the connect failed and start_arg is not NULL, then
 *	start_arg will be executed as the command to start the
 *	server.  After which a connect attempt will be made again.
 *
 *	Returns a pointer to a YiffConnection structure or NULL on
 *	error.  To disconnect and deallocate the YiffConnection structure,
 *	pass it to YIFFDisconnect().
 */
YiffConnection *YIFFConnect(char *start_arg, char *con_arg)
{
	int fd;
	char addr[1024];
	int port;

	int retry_count;
	int total_retries = YIFF_CONNECT_RETRIES;
	int yiff_we_started_server = 0;

	char *strptr;
	YiffConnection *con;



	/* The con_arg must contain something. */
	if(con_arg == NULL)
	    con_arg = DEF_CONNECT_ARG;


	/* Parse con_arg to get addr and port. */
	strncpy(addr, con_arg, 1024);
	addr[1024 - 1] = '\0';

	strptr = strchr(addr, ':');
	if(strptr != NULL)
	    *strptr = '\0';

	port = atoi(strptr + 1);


	/* ********************************************************* */

	/* Connect. */	
	fd = YIFF_CONNECT_TO(addr, port);
	if(fd < 0)
	{
	    /*   Connect failed, implying that the server may not
             *   be running.  Run the server and try to connect again.
	     */

	    /* Check if start_arg is NULL. */
	    if(start_arg == NULL)
		return(NULL);	/* Can't connect, can't start, give up. */
	    else
	        YIFF_START_SERVER(start_arg);

	    /* Try reconnecting again for total_retries times. */
	    for(retry_count = 0; retry_count < total_retries; retry_count++)
	    {
	        fd = YIFF_CONNECT_TO(addr, port);
                if(fd > -1)
	        {
		    /* Server was restarted and we are able to connect. */
		    yiff_we_started_server = 1;
		    break;
		}
		sleep(1);
	    }

	    /* Could not establish connection. */
	    if(fd < 0)
		return(NULL);
	}


        /* Allocate YiffConnection structure. */
        con = (YiffConnection *)calloc(1, sizeof(YiffConnection));
        if(con == NULL)
            return(NULL);   


	/* Set up connection values. */
	con->fd = fd;
	con->we_started_server = yiff_we_started_server;
	con->queued_yevent_pos = -1;
	con->prev_generated_yid = YIDNULL;


	return(con);
}


/*
 *	Changes the attributes audio mode such as;
 *	sample size, bits per second, and sample rate.
 *
 *	You are strongly encuraged to NOT use this function if
 *	at all possible as there is no way to predict if the values
 *	will work, instead use YiffSetModePreset() whenever possible.
 *
 *	Returns -1 on failure, and 0 on success.
 */
int YIFFSetMode(            
        YiffConnection *con,
        int sample_size,	/* 8 or 16. */
        int channels,		/* 1 or 2. */
        int sample_rate,	/* In hz. */
	int direction,		/* 0 for play, 1 for record. */
        int allow_fragmenting,
        int num_fragments,
        int fragment_size	/* In bytes. */
)
{
	char sndbuf[512];
	int len;
	int bytes_written;


	if(con == NULL)
	    return(-1);
	if(con->fd < 0)
	    return(-1);


        sprintf(sndbuf, "%s %i %i %i %i %i %i %i\n",
            YNET_CMD_CHANGEMODE,
            sample_size,
            channels,
            sample_rate,
            direction,
            allow_fragmenting,
            num_fragments,
            fragment_size
        );

        len = strlen(sndbuf);
        bytes_written = send(con->fd, sndbuf, len, 0);
        if(bytes_written < 0)
            return(-1);
        else
            return(0);
}


/*
 *      Changes the audio mode by preset name.
 *
 *      Returns -1 on failure, and 0 on success.
 */
int YIFFSetModePreset(
        YiffConnection *con,
        char *name
)
{
	char lname[256];
        char sndbuf[512];
        int len;
        int bytes_written;   
        
        
        if(con == NULL)
            return(-1);
        if(con->fd < 0)
            return(-1);
	if(name == NULL)
	    return(-1);


	strncpy(lname, name, 256);
	lname[256 - 1] = '\0'; 
        
        sprintf(sndbuf, "%s %s\n",
            YNET_CMD_CHANGEMODEPRESET,
            lname
        );


        len = strlen(sndbuf);
        bytes_written = send(con->fd, sndbuf, len, 0);
        if(bytes_written < 0)
            return(-1);
        else
            return(0);
}


/*
 *	Set cycle.
 *
 *	Note: Intended for configuration program use.
 */
int YIFFSetCycle(
	YiffConnection *con,
	long usec
)
{
        char sndbuf[512];
        int len;
        int bytes_written;

        
        if(con == NULL)
            return(-1);
        if(con->fd < 0)
            return(-1);


        sprintf(sndbuf, "%s %ld\n",
            YNET_CMD_CYCLE,
            usec
        );

        len = strlen(sndbuf);
        bytes_written = send(con->fd, sndbuf, len, 0);
        if(bytes_written < 0)
            return(-1);
        else
            return(0);
}


/*
 *	Syncs with the sound server if block is set to True,
 *	or has just the sound server sync with audio if block is set
 *	to False
 */
void YIFFSync(
	YiffConnection *con,
	Boolean block 
)
{
        char sndbuf[512];
        int len;
        int bytes_written;


        if(con == NULL)
            return;
        if(con->fd < 0)
            return;


        sprintf(sndbuf, "%s\n",
            YNET_CMD_SYNC
        );

        len = strlen(sndbuf);
        bytes_written = send(con->fd, sndbuf, len, 0);
        if(bytes_written < 0)
            return;


	/* Wait for sync. */
	while(block)
	{
/* Need to code this in someday. */
break;
	}


	return;
}


/*
 *	Sends a request sys stats to the server.
 *
 *	Client program must check for sys stats event afterwards.
 */
int YIFFRequestStats(YiffConnection *con)
{
        char sndbuf[512];

        int len;
        int bytes_written;

  
        if(con == NULL)
            return(-1);
        if(con->fd < 0)
            return(-1);


        sprintf(sndbuf, "%s\n",
            YNET_CMD_SYSSTATS
        );

        len = strlen(sndbuf);
        bytes_written = send(con->fd, sndbuf, len, 0);
        if(bytes_written < 0)
            return(-1);
        else
            return(0);
}



/*
 *	Add or remove host.  If op is set to 0, then address
 *	is removed from the hosts list. If op is set to 1, then
 *	address is added to the hosts list.
 */
int YIFFHost(
	YiffConnection *con,
	int op,
	char *address
)
{
        char sndbuf[512];
  
        int len;
	int bytes_written;


        if(con == NULL)
            return(-1);
        if(con->fd < 0)
            return(-1);
	if(address == NULL)
	    return(-1);


	sprintf(sndbuf, "%s %i %s\n",
            YNET_CMD_HOST,
	    op,
	    address
        );

        len = strlen(sndbuf);
        bytes_written = send(con->fd, sndbuf, len, 0);
        if(bytes_written < 0)
            return(-1);
 	else
	    return(0);
}


/*
 *	Sets mixer values.
 */
int YIFFSetMixer(
	YiffConnection *con,
	int mixer_code,
	double value1,
	double value2,
	double value3,
	double value4,
	double value5 
) 
{
        char sndbuf[512];

        int len;
        int bytes_written;   


        if(con == NULL)
            return(-1);
        if(con->fd < 0)
            return(-1);


        sprintf(sndbuf, "%s %i %.6f %.6f %.6f %.6f %.6f\n",
            YNET_CMD_MIXER,
            mixer_code,
            value1,
	    value2,
            value3,
            value4,
            value5
        );


        len = strlen(sndbuf);
        bytes_written = send(con->fd, sndbuf, len, 0);
        if(bytes_written < 0)
            return(-1);
        else
            return(0);
}



/*
 *	Sends a play sound request to the server.
 *	Returns the YID of the new playstack or YIDNULL on failure.
 */
YID YIFFPlaySound(
        YiffConnection *con,
        char *path,		/* Path on disk on server's computer. */
        YDataPosition pos,
        Coefficient left_volume,
        Coefficient right_volume,
	int repeats
)
{
	static YID yid;
        static struct timeval timeout;
        static fd_set writefds;

        static int len, bytes_written;
	static char sndbuf[PATH_MAX + NAME_MAX + 80];


	if((con == NULL) ||
           (path == NULL)
	)
	    return(YIDNULL);

	if(con->fd < 0)
	    return(YIDNULL);


        timeout.tv_sec = YIFF_WRITE_TIMEOUT_SEC;
        timeout.tv_usec = 0;
        FD_ZERO(&writefds);
        FD_SET(con->fd, &writefds);
        select(con->fd + 1, NULL, &writefds, NULL, &timeout);

        if(!FD_ISSET(con->fd, &writefds))
            return(YIDNULL);

	/* Sanitize pos. */
	if(pos < 0)
	    pos = 0;

	/* Increment yid dispense number and get new yid. */
	con->prev_generated_yid++;
	yid = con->prev_generated_yid;

	sprintf(sndbuf, "%s %ld %s %ld %.4f %.4f %i\n",
	    YNET_CMD_PLAY,
	    yid,
	    path,
	    pos,
	    left_volume,
	    right_volume,
            repeats
	);

	len = strlen(sndbuf);
        bytes_written = send(con->fd, sndbuf, len, 0);
	if(bytes_written < 0)
	    return(YIDNULL);
	else
	    return(yid);
}


/*
 *	Sends a YNET_CMD_KILL on yid to the server.
 *
 *	If the yid on the server is still allocated and playing,
 *	it will be destroyed.
 */
void YIFFDestroyPlay(YiffConnection *con, YID yid)
{
        static struct timeval timeout;
        static fd_set writefds;
            
        static int len, bytes_written;
        static char sndbuf[1024];


	if((con == NULL) ||
           (yid == YIDNULL)
	)
	    return;

        if(con->fd < 0)
            return;


        timeout.tv_sec = YIFF_WRITE_TIMEOUT_SEC;
        timeout.tv_usec = 0;
        FD_ZERO(&writefds);
        FD_SET(con->fd, &writefds);
        select(con->fd + 1, NULL, &writefds, NULL, &timeout);

        if(!FD_ISSET(con->fd, &writefds))
            return;


        sprintf(sndbuf, "%s %ld\n",
            YNET_CMD_KILL,
            yid
        );

        len = strlen(sndbuf);
        bytes_written = send(con->fd, sndbuf, len, 0);


	return;
}


/*
 *	Disconnects from YIFF server and frees all allocated
 *	resources.
 *
 *      The pointer con may no longer be used after a call to this
 *      function.
 */
void YIFFDisconnect(YiffConnection *con, Boolean no_shutdown)
{
	char sndbuf[256];


	if(con == NULL)
	    return;


        /* If we started the server, then we should shut it down. */
        if(con->we_started_server &&
           !no_shutdown
        )
        {
	    sprintf(sndbuf, "%s\n",
		YNET_CMD_SHUTDOWN
	    );
            YIFF_SEND_DATA(con->fd, sndbuf);
        }
	else
        {
            sprintf(sndbuf, "%s\n",
                YNET_CMD_DISCONNECT
            );
            YIFF_SEND_DATA(con->fd, sndbuf);
        }


	/* Close connection. */
	if(con->fd > -1)
	{
	    close(con->fd);
	    con->fd = -1;
	}


	/* Free YiffConnection con. */
	free(con);
	con = NULL;


	return;
}


/*
 *	Shuts down the YIFF server, thus disconnecting and
 *	freeing all allocated resources.
 *
 *	The pointer con may no longer be used after a call to this
 *	function.
 */
void YIFFShutdown(YiffConnection *con)
{
        char sndbuf[256];


        if(con == NULL)
            return;


	/* Send shutdown command. */
        sprintf(sndbuf, "%s\n",
            YNET_CMD_SHUTDOWN
        );
        YIFF_SEND_DATA(con->fd, sndbuf);

        /* Close connection. */
        if(con->fd > -1)
        {
            close(con->fd);
            con->fd = -1;
        }


        /* Free YiffConnection con. */
        free(con);
        con = NULL;


	return;
}



/*
 *	Get next event from the server and store it on
 *	the pointer to the YiffEvent event.
 *
 *	If block is set to True, then execution will be
 *	blocked untill next event is recieved.
 */
int YIFFGetNextEvent(
	YiffConnection *con,
	YiffEvent *event,
	Boolean block
)
{
        struct timeval timeout;
        fd_set readfds;

	char buf[RECV_BUF_LEN];
	int i, n, h, status, bytes_read;

	char **strv;
	int strc;

        char **argv;
        int argc;

	YiffEvent *event_ptr;


        if((con == NULL) ||
           (event == NULL)
        )
            return(0);
        
        if(con->fd < 0)
            return(0);

	/* Are there still queued events to be reported? */
	if(con->queued_yevent_pos >= 0)
	{
	    if(con->queued_yevent_pos >= MAX_QUEUED_YIFF_EVENTS)
		con->queued_yevent_pos = MAX_QUEUED_YIFF_EVENTS - 1;

	    event_ptr = &(con->queued_yevent[con->queued_yevent_pos]);

	    YIFF_COPY_EVENTS(event, event_ptr);

	    con->queued_yevent_pos--;


	    return(1);
	}

	/* ********************************************************* */        

	/*   If we should not block, then check if there is data
         *   to be read.
	 */
	if(!block)
	{
	    /* Do not block, check if there is any data to be read. */
            timeout.tv_sec = 0;
            timeout.tv_usec = 0;
            FD_ZERO(&readfds);
            FD_SET(con->fd, &readfds);
            status = select(con->fd + 1, &readfds, NULL, NULL, &timeout);
	    if(status == -1)
		fprintf(stderr, "Select error.\n");

	    if(!FD_ISSET(con->fd, &readfds))
                return(0);
	}

	bytes_read = recv(con->fd, buf, RECV_BUF_LEN, 0);
	if(bytes_read <= 0)
	    return(0);

	/* Sanitize bytes_read and null terminate buf. */
	if(bytes_read >= RECV_BUF_LEN)
	    bytes_read = RECV_BUF_LEN - 1;
	buf[bytes_read] = '\0';


	/* ******************************************************* */
	/* Parse incoming data. */

	/* Explode multiple commands. */
	strv = YIFF_STR_CHAR_EXPLODE(buf, '\n', &strc);
	if((strv == NULL) || (strc <= 0))
	    return(0);

	/* Parse each command segment. */
	for(i = 0; i < strc; i++)
	{
	    if(strv[i] == NULL)
		continue;

	    /* Explode command into arguments. */
	    argv = YIFF_STR_EXPLODE(strv[i], &argc);
	    if((argv == NULL) || (argc <= 0))
		continue;

            /* Get event pointer. */
            if(i <= 0)
            {
		event_ptr = event;
            }
            else
            {
                con->queued_yevent_pos++;
        
                if(con->queued_yevent_pos < 0)
                    con->queued_yevent_pos = 0;
                if(con->queued_yevent_pos >= MAX_QUEUED_YIFF_EVENTS)
                    con->queued_yevent_pos = MAX_QUEUED_YIFF_EVENTS - 1;

		event_ptr = &(con->queued_yevent[con->queued_yevent_pos]);
            }

	    /*   Handle by event type, first argument is event
	     *   type.
	     */

	    /* Change mode. */
	    if(!strcmp(argv[0], YNET_CMD_CHANGEMODE))
	    {
		event_ptr->type = YEventModeChange;
                event_ptr->mode.preset = False;
                event_ptr->mode.mode_name[0] = '\0';

/* Need to know sample size, etc... */

	    }
	    /* Change mode preset. */
	    else if(!strcmp(argv[0], YNET_CMD_CHANGEMODEPRESET))
            {
		event_ptr->type = YEventModeChange;
                event_ptr->mode.preset = True;
		strncpy(
		    event_ptr->mode.mode_name,
		    (argc >= 2) ? argv[1] : "",
		    YMODE_NAME_MAX
		);
		event_ptr->mode.mode_name[YMODE_NAME_MAX - 1] = '\0';
            }
            /* Cycle change. */
            else if(!strcmp(argv[0], YNET_CMD_CYCLE))
            {
		event_ptr->type = YEventCycleChange;
		event_ptr->cycle.cycle_us = (argc >= 2) ?
		    atol(argv[1]) : 0;
            }
            /* Disconnect. */
            else if(!strcmp(argv[0], YNET_CMD_DISCONNECT))
            {
                event_ptr->type = YEventDisconnect;
                event_ptr->disconnect.reason = 0;
            }
            /* Hosts allowed access. */
            else if(!strcmp(argv[0], YNET_CMD_HOST))
            {
		event_ptr->type = YEventHost;
		event_ptr->host.op = (argc >= 2) ? atoi(argv[1]) : 0;

                strncpy(
                    event_ptr->host.address,
                    (argc >= 3) ? argv[2] : "",
                    YHOST_NAME_MAX
                );
                event_ptr->host.address[YHOST_NAME_MAX - 1] = '\0';
            }
            /* Sound killed. */
            else if(!strcmp(argv[0], YNET_CMD_KILL))
            {
                event_ptr->type = YEventSoundKill;
		event_ptr->kill.yid = (argc >= 2) ? atol(argv[1]) : YIDNULL;
            }
            /* Mixer change. */
            else if(!strcmp(argv[0], YNET_CMD_MIXER))
            {
                event_ptr->type = YEventMixer;
                for(h = 0; h < YMIXER_MAX_VALUES; h++)
		{
		    event_ptr->mixer.value[h] =
			(argc >= (h + 2)) ? atof(argv[h + 1]) : 0;
		}
            }
            /* Play started. */
            else if(!strcmp(argv[0], YNET_CMD_PLAY))
            {
                event_ptr->type = YEventSoundPlay;
                event_ptr->play.yid = (argc >= 2) ? atol(argv[1]) : YIDNULL;
            }
            /* Shutdown. */
            else if(!strcmp(argv[0], YNET_CMD_SHUTDOWN))
            {
                event_ptr->type = YEventShutdown;
		event_ptr->shutdown.reason = 0;
            }
            /* Sync completed. */
            else if(!strcmp(argv[0], YNET_CMD_SYNC))
            {
                event_ptr->type = YEventSync;
                event_ptr->sync.cycle_ahead_us = (argc >= 2) ?
		    atol(argv[1]) : 0;
            }
            /* Server system stats. */
            else if(!strcmp(argv[0], YNET_CMD_SYSSTATS))
            {
                event_ptr->type = YEventSystemStats;

                event_ptr->sysstats.cycle_us = (argc >= 2) ?
		    atol(argv[1]) : 0;
                event_ptr->sysstats.compensated_cycle_us = (argc >= 3) ?
                    atol(argv[2]) : 0;
                event_ptr->sysstats.num_fragments = (argc >= 4) ?
                    atoi(argv[3]) : 0;
                event_ptr->sysstats.buf_len = (argc >= 5) ?
                    atoi(argv[4]) : 0;
                event_ptr->sysstats.channels = (argc >= 6) ?
                    atoi(argv[5]) : 0;
                event_ptr->sysstats.sample_size = (argc >= 7) ?
                    atoi(argv[6]) : 0;
                event_ptr->sysstats.sample_rate = (argc >= 8) ?
                    atoi(argv[7]) : 0;
                event_ptr->sysstats.start_time = (argc >= 9) ?
                    atol(argv[8]) : 0;
                event_ptr->sysstats.cycle_load = (argc >= 10) ?
                    atof(argv[9]) : 0;
            }


	    /* Free allocated arguments. */
	    for(n = 0; n < argc; n++)
		free(argv[n]);
	    free(argv);
	    argv = NULL;
	    argc = 0;
	}

	/* Free allocated commands. */
	for(i = 0; i < strc; i++)
	    free(strv[i]);
	free(strv);
	strv = NULL;
	strc = 0;


	return(1);
}
